/*
 * record: record a DVB-T programme
 *
 * Greg Lehey, 6 August 2006
 *
 * Copyright (c) 2006 by Greg Lehey
 *
 *  This software is distributed under the so-called ``Berkeley
 *  License'':
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties
 * of merchantability and fitness for a particular purpose are
 * disclaimed.	In no event shall Greg Lehey be liable for any direct,
 * indirect, incidental, special, exemplary, or consequential damages
 * (including, but not limited to, procurement of substitute goods or
 * services; loss of use, data, or profits; or business interruption)
 * however caused and on any theory of liability, whether in contract,
 * strict liability, or tort (including negligence or otherwise)
 * arising in any way out of the use of this software, even if advised
 * of the possibility of such damage.
 *
 * $Id: record.c,v 1.15 2007/04/04 02:44:26 grog Exp $
 */

/*
 * TO DO:
 *
 * Catch signals, print messages.
 * Get proc title right; it's not clear how to do this with Linux.
 */

#ifdef linux
#define _GNU_SOURCE
#endif

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "timefuncs.h"

#define MUCHSPACE 1048576
/* On some systems this may not be defined */
#ifndef PATHMAX
#define PATHMAX 1024
#endif

void usage (const char *name)
{
  fprintf (stderr,
           "Usage:\n"
           "\n"
           "%s [options] programme filnename [-s date [-e date | -t duration]]\n"
           "\n"
           "-s date\tStart recording at date\n"
           "-e date\tStop recording at date\n"
           "-d\tTuner device number (default 0)\n"
           "-t<time>\tRecord for time specified\n"
           "-n:\tPretend mode: stop after checking parameters\n"
           "-a:\tUse an analogue tuner (otherwise DVB)\n"
           "-q:\tBe quiet: don't show what I'm doing\n"
           "\n"
           "<time> is of the form hh:mm[:ss]\n"
           " date can be [day month-name [year]] hh:mm[:ss]\n"
           "\tor [year-month-day] hh:mm[:ss]\n"
           "Specify either end time or duration, not both\n",
           name );
}

int verbose = 1;					    /* print info as we go */
int analogue = 0;					    /* use an analogue tuner */
int pretending;						    /* parse, print, then stop */
time_t default_record_time = 3600;			    /* record 1 hour by default */
time_t starttime;					    /* time to start recording */
time_t endtime;						    /* time to stop recording */
time_t recordtime;					    /* elapsed time of recording */
time_t last_report;					    /* for displaying progress */
char infilename [PATHMAX];				    /* input file (DVB card) */
char outfilename [PATHMAX];				    /* name of file to write to */
char channelname [PATHMAX];				    /* name of channel */
char channelnick [PATHMAX];				    /* abbreviation used */
char tzap [PATHMAX] = "/usr/local/bin/tzap";		    /* tzap program name */
pid_t mytzap;						    /* pid of our child process */
int infile;						    /* file handle of intput file */
int outfile;						    /* handle of output file */
int overflows;						    /* number of overflow errors on input */
char progname [PATHMAX];				    /* save program name */
char *statusline;					    /* and use this (really argv [0]) to display status */
int statuslength;					    /* length avaiable for status line */
int stopping = 0;					    /* stop at the right place if set */
char iobuf [MUCHSPACE];					    /* transfer MPEG data here */

/*
 * DVB channel list.  The first parameter (name) is a name by which to refer to
 * the channel.  The second name is the official name published by the channel
 * and recognized by tzap.
 *
 * The intention is to use the short names; to avoid confusion, the official
 * names can also be used.
 */
struct channel
{
  char *name;                                               /* our name for the channel */
  char *handle;                                             /* the name that TZAP expects */
  char *description;                                        /* comments about the channel */
  int bps;						    /* bits per second */
}
dvbchannels [] = {
/* first some easy abbreviations */
  {"ABCH", "ABC HDTV", "ABC TV, high definition, no sound", 3750},
  {"ABC2", "ABC2", "ABC TV 2, normal definition", 0},
  {"ABC", "ABC TV", "ABC TV, normal definition", 0},
  {"7", "7 Digital", "7, normal definition", 0},
  {"7H", "7 HD Digital", "7, high definition, no sound", 3750},
  {"9", "NINE Digital", "9, normal program", 0},
  {"9H", "NINE HD", "9, high definition.  Problems?", 3750},
  {"10", "TEN Digital", "10 programm, 704x575", 0},
  {"10H", "TEN HD", "Ten High density 1440x1088, no sound", 3750},
  {"SBSH", "SBS HD", "SBS TV, high definition", 3750},
  {"SBS", "SBS DIGITAL 1", "SBS TV, normal definition", 0},
  {"SBS2", "SBS DIGITAL 2", "SBS news channel, normal definition", 0},
/* and now the "official" names */
  {"ABC HDTV", "ABC HDTV", "", 3750},
  {"ABC TV Adelaide", "ABC TV Adelaide", "", 0},
  {"ABC2", "ABC2", "", 0},
  {"ABC TV", "ABC TV", "", 0},
  {"ABC DiG Radio", "ABC DiG Radio", "", 0},
  {"ABC DiG Jazz", "ABC DiG Jazz", "", 0},
  {"7 Digital", "7 Digital", "", 0},
  {"7 Digital 1", "7 Digital 1", "", 0},
  {"7 Digital 2", "7 Digital 2", "", 0},
  {"7 Digital 3", "7 Digital 3", "", 0},
  {"7 HD Digital", "7 HD Digital", "", 0},
  {"7 Guide", "7 Guide", "", 0},
  {"NINE Digital", "NINE Digital", "", 0},
  {"NINE HD", "NINE HD", "", 0},
  {"TEN Digital", "TEN Digital", "", 0},
  {"TEN Digital 1", "TEN Digital 1", "", 0},
  {"TEN Digital 2", "TEN Digital 2", "", 0},
  {"TEN Digital 3", "TEN Digital 3", "", 0},
  {"TEN Digital", "TEN Digital", "", 0},
  {"TEN Guide", "TEN Guide", "", 0},
  {"TEN HD", "TEN HD", "", 0},
  {"TEN Guide", "TEN Guide", "", 0},
  {"SBS HD", "SBS HD", "", 0},
  {"SBS DIGITAL 1", "SBS DIGITAL 1", "", 0},
  {"SBS DIGITAL 2", "SBS DIGITAL 2", "", 0},
  {"SBS EPG", "SBS EPG", "", 0},
  {"SBS RADIO 1", "SBS RADIO 1", "", 0},
  {"SBS RADIO 2", "SBS RADIO 2", "", 0}
};

/* Print the channel list */
void printchannels ()
{
  int i;

  printf ("\n\nChannel\t\tFull name\tMB/h\tDescription\n\n");
  for (i = 0; i < (sizeof (dvbchannels) / sizeof (struct channel)); i++)
    printf ("%-15s\t%-15s\t%4d\t%s\n",
            dvbchannels [i].name,
            dvbchannels [i].handle,
            dvbchannels [i].bps,
            dvbchannels [i].description);
}

/* Convert channel spec to a name that tzap can use */
void getchannel (const char *channel)
{
  int i;

  strcpy (channelnick, channel);			    /* for reference later */
  for (i = 0; i < (sizeof (dvbchannels) / sizeof (struct channel)); i++)
  {
    if (strcasecmp (channel, dvbchannels [i].name) == 0)    /* found it */
    {
      strcpy (channelname, dvbchannels [i].handle);	    /* put in official name */
      return;
    }
  }
  /* Fall through: didn't find it */
  fprintf (stderr, "Can't find channel %s\n", channel);
  printchannels ();					    /* show what he can use */
  exit (1);
}

/* Signal handler to bury our tzap */
void bury_child (int signum, siginfo_t *info, void *context)
{
  if ((signum == SIGCHLD)				    /* which it has to be */
      && (info->si_pid == mytzap) )			    /* which it also has to be */
  {
    fprintf (stderr,
	     "\ntzap process %d exited, status %d\n",
	     info->si_pid,
	     info->si_status );
    stopping = 1;					    /* stop after flushing buffers */
  }
  else
  {
    if (info->si_signo < 32)                                /* XXX how do we know the size of sys_sigliist ? */
      fprintf (stderr,
               "\nUnexpected signal %s (%d) from process %d\n",
               sys_siglist [info->si_signo],
               info->si_signo,
               info->si_pid );
    else
      fprintf (stderr,
               "\nUnexpected, invalid signal %d from process %d\n",
               info->si_signo,
               info->si_pid );
  }
  exit (1);                                                 /* can't live without our tzap */
}


int main (int argc, char *argv [], char *envp [])
{
  int error;
  int i;
  time_t now;
  int lastreport;
  time_t delay;
  int wait_status;
  struct sigaction sigchild;				    /* disposition for child deaths */
  char *channelid;

  /* This stuff for conversions.  We should be cleverer.  FIXME */
  char starttimebuf [80];
  char endtimebuf [80];
  char *env;
#ifdef linux
  /* *sigh* Linux has this distinction between "large files" and small files. */
  struct stat64 ostat;
#else
  struct stat ostat;
#endif
  int adapter = 0;

  channelid = NULL;
  /*
   * For setproctitle, we overwrite the entire args and env pointers.
   * Find out how much space we have.
   */
  for (i = 0;; i++)
  {
    if (envp [i])
      env = envp [i];
    else
      break;
  }
  statusline = argv [0];				    /* display in ps output */
#ifdef DEBUG
  printf ("argv [0]: %p, last env: %p, envp end: %p, length: %d\n",
          argv [0],
          env,
          env + strlen (env),
          statuslength );
#endif
  statuslength = env + strlen (env) - argv [0];

  strcpy (progname, argv [0]);				    /* save program name */
  if (argc == 1)					    /* no parameters, just print help */
  {
    usage (progname);
    printchannels ();
    exit (0);
  }

  starttime = time (NULL);				    /* default to starting now */

  /* Process command line */
  for (i = 1; i < argc; i++)
  {
    if (strcmp (argv [i], "-a") == 0)
      analogue++;
    if (strcmp (argv [i], "-d") == 0)
    {
      if (++i == argc)					    /* no more parms */
      {
        fprintf (stderr, "Missing device number\n");
        usage (progname);
        return 1;
      }
      else
        adapter = atoi (argv [i]);
    }
    else if (strcmp (argv [i], "-e") == 0)
    {
      if (endtime)					    /* done already, */
      {
        fprintf (stderr, "Duplicate end time\n");
        usage (progname);
        return 1;
      }
      if (++i == argc)					    /* no more parms */
      {
        fprintf (stderr, "Missing end time\n");
        usage (progname);
        return 1;
      }
      else
      {
        /*
         * Depending on how it fails, get_date may have incremnted i or not.
         * Save the current parm in case we need to print an error message.
         */
        char *parm = argv [i];

        endtime = get_date (argv, &i, starttime);
        if (endtime < 0)                                    /* error, */
        {
          fprintf (stderr, "Invalid end time: %s\n", parm);
          usage (progname);
          return 1;
        }
      }
    }
    else if (strcmp (argv [i], "-n") == 0)
    {
      pretending = 1;
    }
    else if (strcmp (argv [i], "-q") == 0)
      verbose = 0;
    else if (strcmp (argv [i], "-s") == 0)
    {
      if (++i == argc)					    /* no more parms */
      {
        fprintf (stderr, "Missing start time\n");
        usage (progname);
        return 1;
      }
      else
      {
        /*
         * Depending on how it fails, get_date may have incremnted i or not.
         * Save the current parm in case we need to print an error message.
         */
        char *parm = argv [i];

        starttime = get_date (argv, &i, starttime);
        if (starttime < 0)                                  /* error, */
        {
          fprintf (stderr, "Invalid start time: %s\n", parm);
          usage (progname);
          return 1;
        }
      }
    }
    else if (strcmp (argv [i], "-t") == 0)
    {
      if (recordtime)					    /* done already, */
      {
        fprintf (stderr, "Duplicate end time\n");
        usage (progname);
        return 1;
      }
      if (argv [i] [2])
        recordtime = get_time (&argv [i] [2]);
      else if (++i == argc)				    /* no more parms */
      {
        fprintf (stderr, "Missing end time\n");
        usage (progname);
        return 1;
      }
      else
        recordtime = get_time (argv [i]);
    }
    else if (strcmp (argv [i], "-v") == 0)
      verbose++;
    else if (channelid)
    {
      if (*outfilename)					    /* all parms specified already */
      {
        fprintf (stderr,
                 "Extraneous information on startup line: %s\n",
                 argv [i]);
        usage (progname);
        return 1;
      }
      else						    /* copy file name */
        strncpy (outfilename, argv [i], PATHMAX);
    }
    else						    /* must be the channel name */
      channelid = argv [i];                                 /* save it */
  }
  if (analogue)
  {
    sprintf (channelname, "Channel %s", channelid);
    strcpy (infilename, "/dev/cxm0");                       /* name of input file */
  }
  else
  {
    getchannel (channelid);                                 /* decipher channel name */
    /* getchannel exits on error, so if we get back here we're safe */
    sprintf (infilename, "/dev/dvb/adapter%d/dvr0", adapter); /* name of input file */
  }
  if (recordtime)
  {
    if (endtime)					    /* specified both */
    {
      fprintf (stderr, "Both end time and run time specified\n");
      usage (progname);
      return 1;
    }
    endtime = starttime + recordtime;
  }
  else if (endtime)					    /* got an end time */
    recordtime = endtime - starttime;			    /* in case we need to print it */
  else							    /* neither specified: default */
  {
    recordtime = default_record_time;
    endtime = starttime + recordtime;
  }
  /* Prepare the record information for printing in various places */
  strftime (starttimebuf, 80, "%e %B %Y %H:%M:%S", localtime (&starttime));
  strftime (endtimebuf, 80, "%e %B %Y %H:%M:%S", localtime (&endtime));
  if (verbose)
  {
    printf ("Recording channel %s (%s)\n"
            "Start time:\t%s\n"
            "End time:\t%s\n",
            channelnick,
            channelname,
            starttimebuf,
            endtimebuf );
    printf ("Duration:\t%s\n"
            "Output file:\t%s\n",
            sprinthms (recordtime),
            outfilename );
  }
  /* Sanity checks */
  now = time (NULL);
  if (outfilename [0] == 0)				    /* didn't get all parms */
  {
    fprintf (stderr, "Missing parameters\n");
    usage (progname);
    return 1;
  }
  else if (recordtime < 1)
  {
    fprintf (stderr, "Negative record time\n");
    usage (progname);
    return 1;
  }
  else if (endtime < now)
  {
    fprintf (stderr, "End time already reached\n");
    return 1;
  }
  else if (starttime < now)
    fprintf (stderr, "Start time already reached\n");
  if (pretending)
  {
    fprintf (stderr, "\nHa ha, just pretending\n");
    return 0;
  }

  /* Open output file now.  Nobody will be happy if it fails just before recording. */

  /*
   * First, find if the file already exists.
   * If so, create a new filename with the date attached.
   */
#ifdef linux
  error = stat64 (outfilename, &ostat);			    /* check on file */
#else
  error = stat (outfilename, &ostat);			    /* check on file */
#endif
  if ((! error) || (errno != ENOENT))			    /* could be lots of things, but the file exists */
  {
    if (verbose)
    {
      if (error)
        fprintf (stderr, "Can't stat %s: %s (%d)\n", outfilename, strerror (errno), errno);
      fprintf (stderr, "%s exists, recording to ", outfilename);
    }
    strftime (&outfilename [strlen (outfilename)],
              PATHMAX - strlen (outfilename),
              "-%Y-%m-%d",
              localtime (&starttime) );
    if (verbose)
      printf ("%s\n", outfilename);
  }
  outfile = open (outfilename,
#ifdef linux
/* Linux requires this flag to tell it to write more than 2 GB.  Sheesh. */
                  O_LARGEFILE |
#endif
                  O_WRONLY |  O_CREAT | O_TRUNC,
                  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);   /* 644 */
  if (outfile < 0)
  {
    fprintf (stderr, "Can't open %s: %s (%d)\n", outfilename, strerror (errno), errno);
    syslog (LOG_ERR | LOG_DAEMON,
            "Can't open %s: %s (%d)\n",
            outfilename,
            strerror (errno),
            errno);
    kill (mytzap, SIGTERM);				    /* kill our dependent tzap */
    return 1;						    /* and give up */
  }

  /* OK, initialization complete.  Wait for the start time to come. */

  now = time (NULL);					    /* get current time */
  delay = starttime - now;				    /* and delay until we start */
  if (delay > 10)					    /* start time long in the future?  */
  {
    snprintf (statusline,
              statuslength,
              "record %s to %s, %s to %s",
              channelname,
              outfilename,
              starttimebuf,
	      endtimebuf);
    if (verbose)
      printf ("\nWill start recording %s to %s in %s at %s\n",
              channelname,
              outfilename,
              sprinthms (delay - 10),
              starttimebuf );
    sleep (delay - 10);					    /* a bit of slack */
  }
  if (analogue)
  {
    char command [80];
    int status;
    int channel;

    channel = atoi (channelid);
    sprintf (command, "setchannel -m 8 %d", channel);
    status = system (command);
    if (status)                                             /* can't tune */
    {
      fprintf (stderr,
               "Can't tune to channel %d: command exited with status %d\n",
               channel,
               status );
      exit (status);
    }
  }
  else                                                      /* DVB */
  {
    /* XXX fix this when we have more than one tuner */
    system ("killall tzap");				    /* kill any running tzap processes */

    /* Ask for information if our tzap dies */
    sigchild.sa_sigaction = bury_child;
    sigchild.sa_flags = SA_SIGINFO;
    sigemptyset (&sigchild.sa_mask);
    sigaction (SIGCHLD, &sigchild, NULL);

    if ((mytzap = fork ()) == 0)                            /* we're the child, */
    {
      char adapternum [8];
      char *argv [] =
        {
          tzap,
          "-r",						    /* provide output to recorder */
          "-S",						    /* be silent */
          "-a",                                             /* select adapter */
          adapternum,
          channelname,
          NULL
        };

      sprintf (adapternum, "%d", adapter);
      if (verbose)
        fprintf (stderr, "Starting %s -r -S -a %s %s\n", tzap, adapternum, channelname);
      execve (tzap, argv, envp);
      syslog (LOG_ERR | LOG_DAEMON,
              "Failed to execve %s: %s (%d)",
              tzap,
              strerror (errno),
              errno);
      return 127;
    }
    else if (mytzap < 0)                                    /* error */
    {
      syslog (LOG_ERR | LOG_DAEMON,
              "Failed to execve %s: %s (%d)",
              tzap,
              strerror (errno),
              errno);
      return 127;
    }
  }

  /* We're the parent.  Open files and wait the last few seconds */
  infile = open (infilename, O_RDONLY, 0);
  if (infile < 0)
  {
    fprintf (stderr, "Can't open %s: %s (%d)\n", infilename, strerror (errno), errno);
    syslog (LOG_ERR | LOG_DAEMON,
            "Can't open %s: %s (%d)\n",
            infilename,
            strerror (errno),
            errno);
    if (! analogue)
      kill (mytzap, SIGTERM);				    /* kill our dependent tzap */
    return 1;						    /* and give up */
  }

  do
  {
    sleep (1);
    now = time (NULL);					    /* get current time */
    delay = starttime - now;				    /* and delay until we start */
  }
  while (delay > 1);					    /* wait for the time to come by */

  /* OK, time to start.  Copy data from the video device to the output file */
  snprintf (statusline,
            statuslength,
            "\nStarting recording %s to %s (%s from %s to %s)",
            channelnick,
            outfilename,
            sprinthms (recordtime),
            starttimebuf,
            endtimebuf );

  if (verbose)
    fprintf (stderr, "%s\n", statusline);
  syslog (LOG_INFO | LOG_DAEMON,
          "Starting recording %s to %s (%s from %s to %s)\n",
          channelnick,
          outfilename,
          sprinthms (recordtime),
          starttimebuf,
          endtimebuf );

  /* Main loop: copy data to output file */
  do
  {
    int incount;
    int warned = 0;					    /* warn once if we write a short block */

    incount = read (infile, iobuf, MUCHSPACE);
    if (incount < 0)
    {
      /* I've seen this on Linux, but not elsewhere.  And the numbers are different! */
      if (errno == EOVERFLOW)
      {
        if (verbose && (overflows == 1))		    /* second one, report it */
          fprintf (stderr,
		   "\nSecond error reading %s: %s (%d)\n",
		   infilename,
		   strerror (errno),
		   errno);
        overflows++;					    /* and count */
      }
      else
      {
        fprintf (stderr, "\nCan't read %s: %s (%d)\n", infilename, strerror (errno), errno);
        syslog (LOG_ERR | LOG_DAEMON,
                "Can't read %s: %s (%d)\n",
                infilename,
                strerror (errno),
                errno);
        if (! analogue)
          kill (mytzap, SIGTERM);                           /* kill our dependent tzap */
        return 1;					    /* and give up */
      }
    }
    else if (incount > 0)				    /* write what we got */
    {
      int outcount;
      outcount = write (outfile, iobuf, incount);

      if (outcount < 0)
      {
        fprintf (stderr,
                 "\nCan't write to  %s: %s (%d)\n",
                 outfilename,
                 strerror (errno),
                 errno);
        syslog (LOG_ERR | LOG_DAEMON,
                "\nCan't write to %s: %s (%d)\n",
                outfilename,
                strerror (errno),
                errno);
        if (! analogue)
          kill (mytzap, SIGTERM);                           /* kill our dependent tzap */
        return 1;					    /* and give up */
      }
      else if (outcount < incount)			    /* this shouldn't happen.  Warn. */
        if (outcount < 0)
        {
          if (! warned)
          {
            warned = 1;
            fprintf (stderr,
                     "\nWrote %d of %d bytes %s)\n",
                     outcount,
                     incount,
                     outfilename);
            syslog (LOG_ERR | LOG_DAEMON,
                    "\nWrote %d of %d bytes %s)\n",
                    outcount,
                    incount,
                    outfilename);
          }
        }
    }
    if (stopping)					    /* time to stop? */
      return stopping;					    /* it's basically an error if we stop here */

    /* Decide whether to update status */
    now = time (NULL);

    if (now > lastreport)				    /* time to show another progress line? */
    {
#ifdef linux
      fstat64 (outfile, &ostat);			    /* get file size */
#else
      fstat (outfile, &ostat);				    /* get file size */
#endif
      long long ppm;					    /* parts per million complete */
      long long int MB;
      int textoffset;

      ppm = (((long long) (now - starttime) * 1000000) / recordtime);
      MB = ostat.st_size >> 20;				    /* size in MB */
      snprintf (statusline,
                statuslength,
                "recorded %s ",
                sprinthms (now - starttime));		    /* elapsed time */
      textoffset = strlen (statusline);			    /* carry on at end */

      if (ppm > 0)					    /* we can calculate estimates */
        snprintf (&statusline [textoffset],
                  statuslength - textoffset,
                  "of %s (%lld%%)  %lld of %lld MB %s",
                  sprinthms (recordtime),		    /* total record time */
                  ppm / 10000,				    /* percentage, */
                  MB,					    /* MB copied */
                  (MB * 1000000) / ppm,			    /* estimated final size */
                  outfilename );			    /* and name of output file */
      else						    /* no estimae */
        snprintf (&statusline [textoffset],
                  statuslength - textoffset,
                  "of %s (%lld%%)  %lld MB %s",
                  sprinthms (recordtime),		    /* total record time */
                  ppm / 10000,				    /* percentage, */
                  MB,					    /* MB copied */
                  outfilename );			    /* and name of output file */
      if (verbose)
      {
        printf ("\r%s", statusline);
        fflush (stdout);
      }
      lastreport = now;
    }
  }
  while (now < endtime);

  /*
   * Finished.  Log results, stop our tzap process, and exit.
   *
   * Slight twist: give now, not endtime, as the end time.  It should be the
   * same thing, of course, but if something goes wrong, this is more honest.
   */
  strftime (endtimebuf, 80, "%e %B %Y %H:%M:%S", localtime (&now));
  if (verbose)
  {
    /* The \n at the end is to get out of the repeating line above. */
    fprintf (stderr,
             "\nCompleted recording %s to %s (%s from %s to %s)\n",
             channelnick,
             outfilename,
             sprinthms (recordtime),
             starttimebuf,
             endtimebuf );
    if (overflows)
      fprintf (stderr, "%d overflows noted \n", overflows);
  }
  syslog (LOG_INFO | LOG_DAEMON,
          "Completed recording %s to %s (%s from %s to %s)\n",
          channelnick,
          outfilename,
          sprinthms (recordtime),
          starttimebuf,
          endtimebuf );
  if (! analogue)
  {
    kill (mytzap, SIGTERM);				    /* kill our dependent tzap */
    for (i = 0; i < 5; i++)				    /* wait 5 seconds for it to die */
    {
      if (waitpid (mytzap, &wait_status, WNOHANG) == mytzap) /* died, */
        return 0;                                           /* we can exit */
      sleep (1);                                            /* wait a second */
    }
    /*
     * If we get here, our tzap hasn't died after 5 seconds.
     * Try again with a stronger weapon.
     */
    fprintf (stderr, "tzap process %d wouldn't die, SIGKILLING\n", mytzap);
    kill (mytzap, SIGKILL);				    /* whole hog */
    for (i = 0; i < 5; i++)				    /* wait 5 seconds for it to die */
    {
      if (waitpid (mytzap, &wait_status, WNOHANG) == mytzap) /* died, */
        return 0;                                           /* we can exit */
      sleep (1);                                            /* wait a second */
    }
    fprintf (stderr,
             "Even a SIGKILL doesn't work on the tzap process %d, giving up\n",
             mytzap);
    syslog (LOG_INFO | LOG_DAEMON,
            "Even a SIGKILL doesn't work on the tzap process %d, giving up\n",
            mytzap);
    return 1;
  }
  return 0;
}

