/*
 * timefuncs: time conversion functions
 *
 * Greg Lehey, 6 August 2006
 *
 * Copyright (c) 2006 by Greg Lehey
 *
 *  This software is distributed under the so-called ``Berkeley
 *  License'':
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties
 * of merchantability and fitness for a particular purpose are
 * disclaimed.	In no event shall Greg Lehey be liable for any direct,
 * indirect, incidental, special, exemplary, or consequential damages
 * (including, but not limited to, procurement of substitute goods or
 * services; loss of use, data, or profits; or business interruption)
 * however caused and on any theory of liability, whether in contract,
 * strict liability, or tort (including negligence or otherwise)
 * arising in any way out of the use of this software, even if advised
 * of the possibility of such damage.
 *
 * $Id: timefuncs.c,v 1.5 2006/10/31 03:20:13 grog Exp $
 */

/*
 * Known problems:
 *
 * Doesn't handle DST correctly.
 * Doesn't handle overly large fields.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

static char *months [] =
    {
    "january",
    "february",
    "march",
    "april",
    "may",
    "june",
    "july",
    "august",
    "september",
    "october",
    "november",
    "december"
      };

static char *mon [] =
    {
    "jan",
    "feb",
    "mar",
    "apr",
    "may",
    "jun",
    "jul",
    "aug",
    "sep",
    "oct",
    "nov",
    "dec" };

/*
 * Like atoi(), but with checks.
 *
 * The string must contain an optional blank leadin, followed by
 * only digits,
 *
 * On error, return -1.
 */
int Atoi (const char *c)
{
  int result = 0;

  while (*c == ' ')
    c++;
  while (isdigit (*c))
    result = result * 10 + *c++ - '0';
  if (*c)                                                   /* trailing crap */
    return -1;
  return result;
}

/*
 * Convert a time, possibly abbreviated, into time_t.
 * If the time isn't completely specific, make it relative to base.
 *
 * timestring is a pointer to a number of strings, such as might come from a
 * program startup argv[], It has the following syntax:
 *
 * [day month-name [year]] hh:mm[:ss]
 * [year-month-day] hh:mm[:ss]
 *
 * Return valid is a time_t for the date, or -1 on error.
 */

time_t get_date (char **argv, int *arg, time_t base)
{
  struct tm mytime;
  struct tm *basetm;                                        /* base time in struc tm format */
  int fields;
  const char *monthname;
  int i;
  const char *hms;
  int defaultdate = 0;                                      /* 1 if no day specified */
  time_t result;

  /*
   * Unfortunately this means that we can't put in user-specified
   * format strings.  We have two formats:
   *
   * 1: (eg 6 August [2006]  11:30[:09])
   *
   * Parm   Meaning
   * 1	    day
   * 2	    month
   * 3      year (may be omitted)
   * 3/4    H:M[:S]
   *
   * 2: (eg 2006-08-06 11:30:09)
   */

  if (argv [*arg] == NULL)                                  /* no parameters, */
    return -1;                                              /* that's got to be invalid */
  memset ((void *) &mytime, 0, sizeof (mytime));            /* clear to start */
  mytime.tm_isdst = -1;                                     /* let mktime work it out */
  if (strchr (argv [*arg], ':') == NULL)                    /* first parm is not time */
  {
    if (strchr (argv [*arg], '-') == NULL)                  /* not yyyy-mm-dd */
    {
      /* XXX consider distinguishing other formats here */
      /* 6 August [2006]  11:30[:09]) */
      mytime.tm_mday = Atoi (argv [*arg]);                  /* day of month or -1 on error */
      if (mytime.tm_mday < 0)
        return mytime.tm_mday;
      /* month */
      monthname = argv [++*arg];
      if (monthname == NULL)                                /* not enough parms */
        return -1;
      if (isalpha (*monthname))
      {
        for (i = 0; i < 12; i++)
          if ((strcasecmp (monthname, months [i])) == 0)
            break;
        if (i == 12)                                        /* nothing found, */
          for (i = 0; i < 12; i++)                          /* try short abbreviations */
            if ((strcasecmp (monthname, mon [i])) == 0)
              break;
        mytime.tm_mon = i;
      }
      else
        mytime.tm_mon = Atoi (monthname) - 1;
      if ((mytime.tm_mon > 11) || (mytime.tm_mon < 0))      /* nothing valid */
        return -1;
      if (argv [++*arg] == NULL)                            /* not enough parms */
        return -1;
      mytime.tm_year = Atoi (argv [*arg]);                  /* suck it and see */
      if (mytime.tm_year < 0)                               /* not a valid year, */
      {
        basetm = localtime (&base);                         /* get base in struct tm format */
        mytime.tm_year = basetm->tm_year;                   /* use this year */
        hms = argv [*arg];                                  /* and reinterpret this value as hms */
      }
      else
        hms = argv [++*arg];                                /* hms is the next field */
    }
    else                                                    /* yyyy-mm-dd hh:mm:ss */
    {
      fields = sscanf (argv [*arg],
                       "%d-%d-%d",
                       &mytime.tm_year,
                       &mytime.tm_mon,
                       &mytime.tm_mday );
      if (fields < 3)                                       /* invalid format */
      {
        fields = sscanf (argv [*arg],
                         "%d-%d",
                         &mytime.tm_mon,
                         &mytime.tm_mday );
        if (fields != 2)                                    /* really crap format */
        {
          fprintf (stderr, "Invalid date format: %s\n", argv [*arg]);
          return -1;
        }
        /* now find out what year this is */
        basetm = localtime (&base);                         /* and in struct tm format */
        mytime.tm_year = basetm->tm_year;
      }
      mytime.tm_mon--;                                      /* months are 0 to 11 */
      hms = argv [++*arg];                                  /* hms is the second field */
    }
    if (mytime.tm_year > 1899)                              /* 4 digits, the way God intended */
      mytime.tm_year -= 1900;                               /* truncate it */
  }
  else                                                      /* only time specified */
  {
    /*
     * We may need to wrap around if hms is less than base.  We do that with the
     * result, because it's easier; here we just note that we have defaulted the
     * date.
     */
    defaultdate = 1;
    /* Get the date from base */
    basetm = localtime (&base);                             /* and in struct tm format */
    mytime.tm_mday = basetm->tm_mday;
    mytime.tm_mon = basetm->tm_mon;
    mytime.tm_year = basetm->tm_year;
    mytime.tm_isdst = basetm->tm_isdst;
    hms = argv [*arg];
  }
  /* Got the date, now get the time */
  if (hms == NULL)                                          /* no hours */
    return -1;                                              /* for the time being, anyway, that's invalid */
  fields = sscanf (hms,
                   "%d:%d:%d",
                   &mytime.tm_hour,
                   &mytime.tm_min,
                   &mytime.tm_sec );
  if (fields < 2)                                           /* accept missing seconds, but only that */
    return -1;

  result = mktime (&mytime);                                /* return as time_t */
  if (defaultdate && (result < base))                       /* wrap around? */
    result += 86400;                                        /* same time tomorrow */
  return result;
}

/*
 * Convert a time in the format hh:mm[:ss] into time_t.  This is a very much
 * simplified version of get_date.
 */
time_t get_time (const char *timestring)
{
  struct tm mytime;
  int fields;

  memset ((void *) &mytime, 0, sizeof (mytime));            /* clear to start */
  fields = sscanf (timestring,
                   "%d:%d:%d",
                   &mytime.tm_hour,
                   &mytime.tm_min,
                   &mytime.tm_sec );
  if (fields < 2)                                           /* invalid format */
  {
    fprintf (stderr, "Invalid time format: %s\n", timestring);
    return -1;
  }
  return mytime.tm_hour * 3600 + mytime.tm_min * 60 + mytime.tm_sec;
}

time_t get_date_old (const char *timestring)
{
  struct tm mytime;
  time_t now;
  struct tm *nowtm;
  int fields;
  char monthname [32];                                      /* store month name here */
  int i;

  /*
   * Try to parse the date.  We don't use strptime(3) because of the
   * pain: first we need to stick the parms back together again, and
   * secondly it craps out if you give it invalid data.
   *
   * Unfortunately this means that we can't put in user-specified
   * format strings.  We have two formats:
   *
   * 1: (eg 6 August 2006  11:30:09)
   *
   * Parm   Meaning
   * 1	    day
   * 2	    month
   * 3      year
   * 4      H:M[:S]
   *
   * 2: (eg 2006-08-06 11:30:09)
   *
   */

  memset ((void *) &mytime, 0, sizeof (mytime));            /* clear to start */
  if (strchr (timestring, '-') == NULL)                     /* not yyyy-mm-dd */
  {
    fields = sscanf (timestring,
                     "%d %s %d %d:%d:%d",
                     &mytime.tm_mday,
                     monthname,
                     &mytime.tm_year,
                     &mytime.tm_hour,
                     &mytime.tm_min,
                     &mytime.tm_sec );
    if (fields < 5)                                         /* invalid format */
    {
      fprintf (stderr, "Invalid date format: %s\n", timestring);
      return -1;
    }
    if (isalpha (*monthname))
    {
      for (i = 0; i < 12; i++)
        if ((strcasecmp (monthname, months [i])) == 0)
          break;
      if (i == 12)                                          /* nothing found, */
        for (i = 0; i < 12; i++)                            /* try short abbreviations */
          if ((strcasecmp (monthname, mon [i])) == 0)
            break;
      mytime.tm_mon = i;
    }
    else
      mytime.tm_mon = atoi (monthname) - 1;
    if ((mytime.tm_mon > 11) || (mytime.tm_mon < 0))        /* nothing valid */
    {
      fprintf (stderr, "invalid month: '%s'\n", monthname);
      return -1;
    }
  }
  else                                                      /* yyyy-mm-dd hh:mm:ss */
  {
    fields = sscanf (timestring,
                     "%d-%d-%d %d:%d:%d",
                     &mytime.tm_year,
                     &mytime.tm_mon,
                     &mytime.tm_mday,
                     &mytime.tm_hour,
                     &mytime.tm_min,
                     &mytime.tm_sec );
    if (fields < 5)                                         /* invalid format */
    {
      fields = sscanf (timestring,
                       "%d-%d %d:%d",
                       &mytime.tm_mon,
                       &mytime.tm_mday,
                       &mytime.tm_hour,
                       &mytime.tm_min);
      if (fields != 4)                                      /* really crap format */
      {
        fprintf (stderr, "Invalid date format: %s\n", timestring);
        return -1;
      }
      /* now find out what year this is */
      now = time (NULL);                                    /* current time */
      nowtm = localtime (&now);                             /* and in struct tm format */
      mytime.tm_year = nowtm->tm_year;
    }
    mytime.tm_mon--;                                        /* months are 0 to 11 */
  }
  if (mytime.tm_year > 1899)                                /* 4 digits, the way God intended */
    mytime.tm_year -= 1900;                                 /* truncate it */
  return mktime (&mytime);                                  /* return as time_t */
}

/* Print a time as hh:mm:ss */
char *sprinthms (time_t when)
{
  static char buf [20];
int h;                                                      /* to convert recording time */
  int m;
  int s;

  h = when / 3600;
  s = when % 3600;
  m = s / 60;
  s = s % 60;

  sprintf (buf, "%02d:%02d:%02d", h, m, s);
  /* Why does the compiler complain without this cast? */
  return (char *) &buf;
}

#ifdef UNIT_TEST
int main (int argc, char *argv [])
{
  time_t date;
  int arg = 1;

  if ((date = get_date (argv, &arg, time (NULL))) == -1)
    puts ("*** Error ***");
  else
    printf ("%s\n", ctime (&date));
  return 0;
}
#endif

